#!/bin/bash
# 
#  Self contained LiveCPD install script.
#    To use, execute like /bin/bash ./livecpd2-installer.sh
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#

REQUIRED_PROGS="perl tar gunzip cat whoami uname mkdir mount readlink chmod"

for prog in $REQUIRED_PROGS; do
    if ! which $prog >/dev/null 2>&1; then
        echo "This script requires ${prog} to be in your PATH" 1>&2
        exit 1
    fi
done

if [ `whoami` != "root" ]; then
    echo "Warning: you will need raw device access for this script to continue."
    read -p " Continue anyway without root access? [y/N] " NOROOT_CONTINUE
    if [ "${NOROOT_CONTINUE}" != "y" ] && [ "${NOROOT_CONTINUE}" != "Y" ]; then
        echo "Aborted" 1>&2
        exit 1
    fi
fi

read -p "Device to install to: [/dev/sda1] " DEVICE_TARGET
if [ -z "${DEVICE_TARGET}" ]; then
    DEVICE_TARGET="/dev/sda1"
fi
DEVICE=`echo ${DEVICE_TARGET} | perl -pe 's/^(\D+)\d*$/$1/'`

if grep -q "^${DEVICE}" /proc/mounts >/dev/null 2>&1; then
    echo "${DEVICE} has one or more currently mounted partitions. Aborted."
    exit 1
fi

read -p "Flag $DEVICE_TARGET as an override device? [y/N] " SETUP_OVERRIDE
if [ "${SETUP_OVERRIDE}" != "y" ] && [ "${SETUP_OVERRIDE}" != "Y" ]; then
    SETUP_OVERRIDE=""
fi


THIS_FILE=`readlink -f $0`
GET_TARBALL="perl -MMIME::Base64 -ne 'if (\$pf) { print decode_base64(\$_); next; } \$pf=1 if /^TARBALLDATASTART$/;' $THIS_FILE"

extract_files_to_dir() {
    DIR=$1
    eval $GET_TARBALL | tar -xOf - image.tar | tar -xof - -C $DIR
    if [ $? -ne 0 ]; then
        return 1
    fi
    chmod a-w $DIR/syslinux.cfg $DIR/bootmsg.txt $DIR/initrd.gz $DIR/vmlinuz $DIR/root.img >/dev/null 2>&1
    if [ ! -z "${SETUP_OVERRIDE}" ]; then
        touch $DIR/override.img
        chmod ug+rw $DIR/override.img >/dev/null 2>&1
    fi
    return 0
}
get_util_file() {
    FILE=$1
    eval $GET_TARBALL | tar -xOf - utils.tar.gz | gunzip - | tar -xOf - $FILE
}


SYSLINUX_BINARY=
SYSLINUX_BINARY_LOCAL=
MKDISKIMAGE_BINARY=
MKDISKIMAGE_BINARY_LOCAL=
MOUNT_POINT="/tmp/livecpd2-dev-mnt-$$"
USBIMG_DIR="/tmp/livecpd2-img-$$"
cleanup() {
    if [ -n "${SYSLINUX_BINARY}" ] && [ -n "${SYSLINUX_BINARY_LOCAL}" ]; then
        rm -f "${SYSLINUX_BINARY}" >/dev/null 2>&1
    fi
    if [ -n "${MKDISKIMAGE_BINARY}" ] && [ -n "${MKDISKIMAGE_BINARY_LOCAL}" ]; then
        rm -f "${MKDISKIMAGE_BINARY}" >/dev/null 2>&1
    fi
}
fail() {
    if [ -n "${SYSLINUX_BINARY}" ] && [ -n "${SYSLINUX_BINARY_LOCAL}" ]; then
        rm -f "${SYSLINUX_BINARY}" >/dev/null 2>&1
    fi
    if [ -n "${MKDISKIMAGE_BINARY}" ] && [ -n "${MKDISKIMAGE_BINARY_LOCAL}" ]; then
        rm -f "${MKDISKIMAGE_BINARY}" >/dev/null 2>&1
    fi
    if [ -d "${MOUNT_POINT}" ]; then
        umount $MOUNT_POINT >/dev/null 2>&1
        rmdir $MOUNT_POINT >/dev/null 2>&1
    fi
    if [ -d "${USBIMG_DIR}" ]; then
        rm -rf $USBIMG_DIR >/dev/null 2>&1
    fi

    echo -e "\n\n*** FAILED ****\n"
    exit 1
}

if [ -d "${DEVICE_TARGET}" ]; then
    echo "${DEVICE_TARGET} looks like a directory."
    echo "  Extracting the device image there.  You must install your own boot loader."
    echo "  See ${DEVICE_TARGET}/syslinux.cfg for kernel parameters."
    set -e
    extract_files_to_dir "${DEVICE_TARGET}"
    exit 0
fi


trap fail INT TERM HUP

install_normal() {
    if [ `uname -i` != "i386" ]; then
        SYSLINUX_BINARY=`which syslinux 2>/dev/null`
        if [ -z "${SYSLINUX_BINARY}" ]; then
            echo "WARNING: This doesn't look like i386 compatible, and you don't have "
            echo "  syslinux installed, this likely means the install can't complete."
            read -p "Try continuing anyway? [y/N] " CONTINUE_ANYWAY
            if [ "${CONTINUE_ANYWAY}" != "y" ] && [ "${CONTINUE_ANYWAY}" != "Y" ]; then
                echo "Aborted"
                exit 1
            fi
        else
            echo "WARNING: This doesn't look like i386 compatible, but it looks like syslinux"
            echo "  is installed, so it will be used."
        fi
    fi

    if [ -z "${SYSLINUX_BINARY}" ]; then
        SYSLINUX_BINARY="/tmp/syslinux-$$"
        SYSLINUX_BINARY_LOCAL=1
        echo -n "Extracting syslinux..."
        get_util_file syslinux > $SYSLINUX_BINARY || fail
        chmod u+x $SYSLINUX_BINARY || fail
        echo "Done"
    fi

    echo -n "Installing bootloader..."
    eval $SYSLINUX_BINARY $DEVICE_TARGET || fail
    echo "Done"

    echo -n "Mounting ${DEVICE_TARGET}..."
    mkdir -p $MOUNT_POINT
    mount -t msdos $DEVICE_TARGET $MOUNT_POINT || fail
    echo "Done"

    echo -n "Extracting to ${DEVICE_TARGET}..."
    extract_files_to_dir $MOUNT_POINT || fail
    echo "Done"

    echo -n "Unmounting ${DEVICE_TARGET}..."
    umount $MOUNT_POINT || fail
    echo "Done"

    echo -n "Cleaning up..."
    rmdir $MOUNT_POINT || fail
    cleanup
    echo "Done"
}

read -p "Use USB-ZIP compatibility mode? (WARNING: DESTRUCTIVE) [y/N] " USE_USBZIP
if [ "${USE_USBZIP}" != "y" ] && [ "${USE_USBZIP}" != "Y" ]; then
    if [ ! -b "${DEVICE_TARGET}" ]; then
        echo "${DEVICE_TARGET} is not a block device." 1>&2
        exit 1
    fi

    read -p "WARNING: This will overwrite the boot sector of ${DEVICE}. Proceed? [y/N] " DO_NORMAL
    if [ "${DO_NORMAL}" != "y" ] && [ "${DO_NORMAL}" != "Y" ]; then
        echo "Aborted"
        exit 1
    fi
    install_normal
    exit 0
fi

if [ ! -b "${DEVICE}" ]; then
    echo "${DEVICE} is not a block device." 1>&2
    exit 1
fi

read -p "*** WARNING *** This will ERASE all data on ${DEVICE}. Continue? [y/N] " USBZIP_CONINTUE
if [ "${USBZIP_CONINTUE}" != "y" ] && [ "${USBZIP_CONINTUE}" != "Y" ]; then
    echo "Aborted" 1>&2
    exit 1
else
    BLOCKDEV=blockdev
    if ! which blockdev >/dev/null 2>&1; then
        if ! -x /sbin/blockdev; then
            echo "This script requires blockdev to be in your PATH" 1>&2
            exit 1
        else
            BLOCKDEV=/sbin/blockdev
        fi
    fi

    MKDISKIMAGE_BINARY=`which mkdiskimage 2>/dev/null` 
    if [ -z "${MKDISKIMAGE_BINARY}" ]; then
        MKDISKIMAGE_BINARY="/tmp/mkdiskimage-$$"
        MKDISKIMAGE_BINARY_LOCAL=1
        echo -n "Extracting mkdiskimage..."
        get_util_file mkdiskimage > $MKDISKIMAGE_BINARY || fail
        chmod u+x $MKDISKIMAGE_BINARY || fail
        echo "Done"
    fi

    echo -n "Regenerating ${DEVICE}..."
    eval $SYSLINUX_BINARY -4 "${DEVICE}" 0 64 32 || fail
    sleep 1
    eval $BLOCKDEV --rereadpt "${DEVICE}" || fail
    echo "Done"

    DEVICE_TARGET="${DEVICE}4"
    install_normal

    exit 0
fi

if /bin/false; then
TARBALLDATASTART
