#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "headers.h"
#include "archive_append.h"

static ssize_t sread(int fd, void *buf, size_t count) {
	ssize_t nrd = 0;
	ssize_t rc;
	while (count > 0) {
		rc = read(fd, (char *)buf+nrd, count);
		if (rc == -1) return -1;
		if (rc == 0) return 0;
		nrd += rc;
		count -= rc;
	}
	return nrd;
}
static ssize_t swrite(int fd, void *buf, size_t count) {
	ssize_t nwr = 0;
	ssize_t rc;
	while (count > 0) {
		rc = write(fd, (char *)buf+nwr, count);
		if (rc == -1) return -1;
		nwr += rc;
		count -= rc;
	}
	return nwr;
}

#define WRITE_EXACT(fd,bfr,cnt)	{ if (swrite(fd,bfr,cnt) != cnt) return -1; }

static int insert_regular_data( int fdout, char *filename, struct stat *sb ) {
	int fdin;
	OVRDA_FileHeader hdr;
	size_t cnt;
	ssize_t nrd;
	char bfr[512];

	/* Write header */
	hdr.mode = (unsigned long)(sb->st_mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX|S_ISGID|S_ISUID));
	hdr.atime = (unsigned long)(sb->st_atime);
	hdr.mtime = (unsigned long)(sb->st_atime);
	hdr.size = (unsigned long)(sb->st_size);
	WRITE_EXACT(fdout, &hdr, sizeof(hdr));
	
	/* Copy file data */
	if ((fdin=open(filename, O_RDONLY)) == -1)
		return -1;
	cnt = (size_t)sb->st_size;
	while (cnt > 0) {
		nrd = cnt > sizeof(bfr) ? sizeof(bfr) : cnt;
		if ((nrd=read(fdin, bfr, nrd)) <= 0)
			return -1;
		WRITE_EXACT(fdout, bfr, nrd);
		cnt -= nrd;
	}
	close(fdin);
	
	return 0;
}

static int insert_dir_data( int fdout, char *filename, struct stat *sb ) {
	OVRDA_DirHeader hdr;

	hdr.mode = (unsigned long)(sb->st_mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX|S_ISGID|S_ISUID));
	WRITE_EXACT(fdout, &hdr, sizeof(hdr));

	return 0;
}

static int insert_symlink_data( int fdout, char *filename, struct stat *sb ) {
	OVRDA_SymlinkHeader hdr;
	char lnkbuff[512];
	int lnklen;

	/* Read link contents */
	if ((lnklen=readlink(filename, lnkbuff, sizeof(lnkbuff))) <= 0)
		return -1;

	/* Write header */
	hdr.ln_len = (unsigned long)lnklen;
	WRITE_EXACT(fdout, &hdr, sizeof(hdr));

	/* Write link contents */
	WRITE_EXACT(fdout, lnkbuff, lnklen);
	
	return 0;
}

static int insert_special_data( int fdout, char *filename, struct stat *sb ) {
	OVRDA_SpecialHeader hdr;

	hdr.mode = (unsigned long)(sb->st_mode);
	hdr.dev = (unsigned long)(sb->st_rdev);
	WRITE_EXACT(fdout, &hdr, sizeof(hdr));

	return 0;
}

int archive_append_file( int fdout, char *filename ) {
	struct stat sb;
	OVRDA_AllHeader hdr;

	/* Stat now so we get the right atime */
	if (lstat(filename, &sb) == -1)
		return -1;

	/* Insert the univeral header */
	if (S_ISREG(sb.st_mode)) {
		hdr.type = OVRDAFT_REG;
	} else if (S_ISDIR(sb.st_mode)) {
		hdr.type = OVRDAFT_DIR;
	} else if (S_ISLNK(sb.st_mode)) {
		hdr.type = OVRDAFT_LINK;
	} else if (S_ISCHR(sb.st_mode) || S_ISBLK(sb.st_mode) || S_ISFIFO(sb.st_mode) || S_ISSOCK(sb.st_mode)) {
		hdr.type = OVRDAFT_SPECIAL;
	} else
		return -1;
	hdr.uid = sb.st_uid;
	hdr.gid = sb.st_gid;
	hdr.fn_len = strlen(filename) + (filename[0] == '/' ? 0 : 1);
	WRITE_EXACT(fdout, &hdr, sizeof(hdr));
	/* Write filename */
	if (filename[0] != '/')
		WRITE_EXACT(fdout, "/", 1);
	WRITE_EXACT(fdout, filename, strlen(filename));

	/* Insert specific type data */
	if (S_ISREG(sb.st_mode)) {
		return insert_regular_data(fdout, filename, &sb);
	} else if (S_ISDIR(sb.st_mode)) {
		return insert_dir_data(fdout, filename, &sb);
	} else if (S_ISLNK(sb.st_mode)) {
		return insert_symlink_data(fdout, filename, &sb);
	} else if (S_ISCHR(sb.st_mode) || S_ISBLK(sb.st_mode) || S_ISFIFO(sb.st_mode) || S_ISSOCK(sb.st_mode)) {
		return insert_special_data(fdout, filename, &sb);
	}

	return -1;
}
