#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <utime.h>

#include "headers.h"
#include "archive_extract.h"

static unsigned int g_total_read = 0;

static ssize_t sread(int fd, void *buf, size_t count) {
	ssize_t nrd = 0;
	ssize_t rc;
	while (count > 0) {
		rc = read(fd, (char *)buf+nrd, count);
		if (rc == -1) return -1;
		if (rc == 0) return 0;
		nrd += rc;
		count -= rc;
	}
	g_total_read += nrd;
	return nrd;
}
static ssize_t swrite(int fd, void *buf, size_t count) {
	ssize_t nwr = 0;
	ssize_t rc;
	while (count > 0) {
		rc = write(fd, (char *)buf+nwr, count);
		if (rc == -1) return -1;
		nwr += rc;
		count -= rc;
	}
	return nwr;
}

#define READ_EXACT(fd,bfr,cnt)	{ if (sread(fd,bfr,cnt) != cnt) { return -1; } }

static int extract_regular( int dowrite, int fdin, OVRDA_AllHeader *ghdr, char *filename ) {
	int fdout = -1;
	OVRDA_FileHeader hdr;
	size_t cnt;
	ssize_t nrd;
	char bfr[1024];
	struct utimbuf ut;

	READ_EXACT(fdin, &hdr, sizeof(hdr));
	if (dowrite) {
		unlink(filename);
		rmdir(filename);
		if ((fdout=open(filename, O_WRONLY|O_CREAT|O_TRUNC)) == -1)
			return -1;
	}

	cnt = hdr.size;

	while (cnt > 0) {
		nrd = cnt > sizeof(bfr) ? sizeof(bfr) : cnt;
		if ((nrd=read(fdin, bfr, nrd)) <= 0)
			return -1;
		g_total_read += nrd;
		if (fdout != -1) {
			if (swrite(fdout, bfr, nrd) != nrd)
				return -1;
		}
		cnt -= nrd;
	}

	if (fdout != -1) close(fdout);

	if (dowrite) {
		if (chmod(filename, hdr.mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX|S_ISGID|S_ISUID)) != 0)
			return -1;
		ut.actime = hdr.atime;
		ut.modtime = hdr.mtime;
		if (utime(filename, &ut) != 0)
			return -1;
	}

	return 0;
}

static int extract_dir( int dowrite, int fdin, OVRDA_AllHeader *ghdr, char *filename ) {
	OVRDA_DirHeader hdr;
	struct stat st;
	
	READ_EXACT(fdin, &hdr, sizeof(hdr));
	if (dowrite) {
		/* If it doesn't exist as a directory, create that */
		if (stat(filename, &st) == -1 || !(S_ISDIR(st.st_mode))) {
			unlink(filename);
			if (mkdir(filename, hdr.mode & (S_IRWXU|S_IRWXG|S_IRWXO)) != 0)
				return -1;
		}
		if (chmod(filename, hdr.mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX|S_ISGID|S_ISUID)) != 0)
			return -1;
	}

	return 0;
}

static int extract_symlink( int dowrite, int fdin, OVRDA_AllHeader *ghdr, char *filename ) {
	OVRDA_SymlinkHeader hdr;
	char lnkbuff[512];
	
	READ_EXACT(fdin, &hdr, sizeof(hdr));
	if (dowrite) {
		if (hdr.ln_len > sizeof(lnkbuff)-1) hdr.ln_len = sizeof(lnkbuff)-1;
		READ_EXACT(fdin, lnkbuff, hdr.ln_len);
		lnkbuff[hdr.ln_len] = '\0';
		unlink(filename);
		rmdir(filename);
		if (symlink(lnkbuff, filename) != 0)
			return -1;
	} else {
		if (lseek(fdin, hdr.ln_len, SEEK_CUR) == (off_t)-1)
			return -1;
	}
	return 0;
}

static int extract_special( int dowrite, int fdin, OVRDA_AllHeader *ghdr, char *filename ) {
	OVRDA_SpecialHeader hdr;
	mode_t mt;
	dev_t dt;
	
	READ_EXACT(fdin, &hdr, sizeof(hdr));
	if (dowrite) {
		if (S_ISCHR(hdr.mode))
			mt = S_IFCHR;
		else if (S_ISBLK(hdr.mode))
			mt = S_IFBLK;
		else if (S_ISSOCK(hdr.mode))
			mt = S_IFIFO;
		else
			mt = S_IFIFO;
		if (S_ISCHR(hdr.mode) || S_ISBLK(hdr.mode))
			dt = hdr.dev;
		else
			dt = 0;
		unlink(filename);
		rmdir(filename);
		if (mknod(filename, mt, dt) != 0)
			return -1;
		if (chmod(filename, hdr.mode & (S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX|S_ISGID|S_ISUID)) != 0)
			return -1;
	}

	return 0;
}

#define NDOTS		25
#define DPERPCT		5
int archive_extract( char *prefix, int fdin, int display, archive_filter filter ) {
	OVRDA_AllHeader hdr;
	char *filename = NULL;
	int fc, rd;
	unsigned int fsize;
	double dlt;
	int didx;

	if (sread(fdin, &fsize, 4) == -1)
		return -1;
	g_total_read = 4;

	if (display) {
		printf("0%%");
		fflush(stdout);
		didx = 1;
	}

	while ((rd=sread(fdin, &hdr, sizeof(hdr))) == sizeof(hdr)) {
		if (!hdr.fn_len) return -1;

		if (display && fsize) {
			dlt = ((double)g_total_read/(double)fsize) * (double)NDOTS;
			while (didx <= (int)(dlt)) {
				if (didx && didx < NDOTS && (didx % DPERPCT) == 0) {
					printf("%d%%", (int)(((double)didx/(double)NDOTS)*100.0));
				} else {
					printf(".");
				}
				fflush(stdout);
				didx++;
			}
		}
		
		filename = realloc(filename, hdr.fn_len+strlen(prefix)+1);
		READ_EXACT(fdin, filename, hdr.fn_len);
		filename[hdr.fn_len] = '\0';
		memmove(filename+strlen(prefix), filename, strlen(filename)+1);
		memcpy(filename, prefix, strlen(prefix));

		if (filter) fc = (*filter)(fdin, &hdr, filename);
		else fc = 0;

		if (fc == 2) continue;

		switch(hdr.type) {
		case OVRDAFT_REG:
			if (extract_regular((fc==0) ? 1 : 0, fdin, &hdr, filename) != 0)
				return -1;
			break;
		case OVRDAFT_DIR:
			if (extract_dir((fc==0) ? 1 : 0, fdin, &hdr, filename) != 0)
				return -1;
			break;
		case OVRDAFT_LINK:
			if (extract_symlink((fc==0) ? 1 : 0, fdin, &hdr, filename) != 0)
				return -1;
			break;
		case OVRDAFT_SPECIAL:
			if (extract_special((fc==0) ? 1 : 0, fdin, &hdr, filename) != 0)
				return -1;
			break;
		default:
			return -1;
		}

		if (fc == 0) {
			if (lchown(filename, hdr.uid, hdr.gid) != 0)
				return -1;
		}
	}

	if (rd == -1)
		return -1;

	if (filename) free(filename);

	if (display) {
		printf("100%%");
		fflush(stdout);
	}

	return 0;
}
